# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/tools/api/generator/create_python_api.py script.
"""Operations for working with string Tensors.

See the @{$python/string_ops} guide.

"""

from __future__ import print_function

from tensorflow.python import regex_full_match

del print_function
