# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/tools/api/generator/create_python_api.py script.
"""SavedModel utility functions.

Utility functions to assist with setup and construction of the SavedModel proto.

"""

from __future__ import print_function

from tensorflow.python.saved_model.utils import build_tensor_info
from tensorflow.python.saved_model.utils import get_tensor_from_tensor_info

del print_function
