# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/tools/api/generator/create_python_api.py script.
"""Public API for tf.estimator.inputs namespace.
"""

from __future__ import print_function

from tensorflow.python.estimator.inputs.inputs import numpy_input_fn
from tensorflow.python.estimator.inputs.inputs import pandas_input_fn

del print_function
